#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include <stdint.h>
#include <math.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "opcode.h"
#include "misc.h"
#include "buffer.h"
#include "vm.h"
#include "vmlibcall.h"
#include "mem_area.h"
#include "buffer.h"
#include "obj.h"
#include "bo.h"

#define	STACK_SIZE	98192

static const char *except_msgs[] = {
	"unknown instruction opcode",
	"unknown area type",
	"exceed stack memory range",
	"exceed data section range",
	"exceed text section range",
	"exceed ro section range",
	"exceed array bounds",
	"division by zero",
	"library call failed",
	"unknown reference point",
	"read from outputs",
	"write to inputs",
	"memory area data type",
	"inputs area memory range",
	"outputs area memory range",
	"bit number"
};

struct vm *vm_create()
{
	struct vm *machine;

	machine = xmalloc0(sizeof(*machine));

	return machine;
}

ret_t vm_destroy(struct vm **machine)
{
	return_val_if_fail(machine != NULL, ret_err);
	return_val_if_fail(*machine != NULL, ret_err);

	xfree(*machine);
	*machine = NULL;

	return ret_ok;
}

ret_t vm_io_area_set(struct vm *machine, void *inputs, unsigned int inputs_sz,
		     void *outputs, unsigned int outputs_sz, int byte_order)
{
	return_val_if_fail(machine != NULL, ret_err);	

	if (inputs != NULL) {
		machine->inputs = buffer_create_static_full(inputs, inputs_sz);
		machine->inputs_sz = inputs_sz;
	}

	if (outputs != NULL) {
		machine->outputs = buffer_create_static_full(outputs, outputs_sz);
		machine->outputs_sz = outputs_sz;
	}

	machine->byte_order = byte_order;

	return ret_ok;
}

ret_t vm_io_area_unset(struct vm *machine)
{
	return_val_if_fail(machine != NULL, ret_err);	

	if (machine->inputs != NULL)
		buffer_free(machine->inputs);

	if (machine->outputs != NULL)
		buffer_free(machine->outputs);

	return ret_ok;
}

static void raise_exception(struct vm *machine, except_type_t type)
{
	machine->state = VM_EXCEPTION;

	if (machine->vm_exception_callback != NULL)
		machine->vm_exception_callback(machine, type, except_msgs[type]);
	else
		xerror(1, "exception: %s\n\t%%pc -- 0x%x %%sp -- 0x%x %%fp -- 0x%x",
			except_msgs[type], machine->pc, machine->sp, machine->fp);
}

static void check_area_bounds(struct vm *machine, unsigned char area, int32_t offs, size_t size)
{
	return_if_fail(machine != NULL);

	switch (area) {
	case DATA_OFFS:
		if (offs < 0 || offs+size > machine->data_size)
			raise_exception(machine, EXCEPT_OO_DATA);
		break;
	case FP_OFFS:
		if (offs < 0 && machine->fp+offs < 0)
			raise_exception(machine, EXCEPT_OO_STACK);
		else if (offs > 0 && machine->fp+offs+size > machine->stack_size)
			raise_exception(machine, EXCEPT_OO_STACK);
		break;
	case IAREA_OFFS:
		if (offs < 0  || offs+size > machine->inputs_sz)
			raise_exception(machine, EXCEPT_OO_INPUTS);
		break;
	case QAREA_OFFS:
		if (offs < 0  || offs+size > machine->outputs_sz)
			raise_exception(machine, EXCEPT_OO_OUTPUTS);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}
}

static uint32_t peek_operand(struct vm *machine)
{
	uint32_t res;
	unsigned char *data;

	data = machine->text + machine->pc;

	res  = data[3];
	res |= data[2] << 8;
	res |= data[1] << 16;
	res |= data[0] << 24;

	return res;
}

void vm_stack_shrink(struct vm *machine, size_t size)
{
	return_if_fail(machine != NULL);

	if (machine->sp < size)
		raise_exception(machine, EXCEPT_OO_STACK);

	machine->sp -= size;
}

void vm_stack_grow(struct vm *machine, size_t size)
{
	return_if_fail(machine != NULL);

	if (machine->sp+size > machine->stack_size)
		raise_exception(machine, EXCEPT_OO_STACK);

	machine->sp += size;
}

int32_t vm_stack_get_int(struct vm *machine)
{
	int32_t res;

	return_val_if_fail(machine != NULL, 0);

	res = machine->stack[machine->sp >> 2];
	vm_stack_grow(machine, 4);

	return res;
}

int32_t vm_stack_peek_int(struct vm *machine)
{
	int32_t res;

	return_val_if_fail(machine != NULL, 0);

	res = machine->stack[machine->sp >> 2];

	return res;
}

void vm_stack_put_int(struct vm *machine, int32_t num)
{
	return_if_fail(machine != NULL);

	vm_stack_shrink(machine, 4);
	machine->stack[machine->sp >> 2] = num;
}

void vm_stack_poke_int(struct vm *machine, int32_t num)
{
	return_if_fail(machine != NULL);

	machine->stack[machine->sp >> 2] = num;
}

float vm_stack_get_float(struct vm *machine)
{
	float res;

	return_val_if_fail(machine != NULL, 0);

	res = *(float *)&machine->stack[machine->sp >> 2];
	vm_stack_grow(machine, 4);

	return res;
}

float vm_stack_peek_float(struct vm *machine)
{
	float res;

	return_val_if_fail(machine != NULL, 0);

	res = *(float *)&machine->stack[machine->sp >> 2];

	return res;
}

void vm_stack_put_float(struct vm *machine, float fl)
{
	return_if_fail(machine != NULL);

	vm_stack_shrink(machine, 4);
	machine->stack[machine->sp >> 2] = *(uint32_t *)&fl;
}

void vm_stack_poke_float(struct vm *machine, float fl)
{
	return_if_fail(machine != NULL);

	machine->stack[machine->sp >> 2] = *(uint32_t *)&fl;
}

char vm_stack_get_char(struct vm *machine)
{
	uint32_t res;

	return_val_if_fail(machine != NULL, 0);

	res = machine->stack[machine->sp >> 2];
	vm_stack_grow(machine, 4);

	return (char)res;
}

inline char vm_stack_peek_char(struct vm *machine)
{
	uint32_t res;

	return_val_if_fail(machine != NULL, 0);

	res = machine->stack[machine->sp >> 2];

	return (char)res;
}

void vm_stack_put_char(struct vm *machine, char ch)
{
	return_if_fail(machine != NULL);

	vm_stack_shrink(machine, 4);
	machine->stack[machine->sp >> 2] = (uint32_t)ch;
}

void vm_stack_poke_char(struct vm *machine, char ch)
{
	return_if_fail(machine != NULL);

	machine->stack[machine->sp >> 2] = (uint32_t)ch;
}

void vm_stack_get_string(struct vm *machine, char dst[STRING_LEN])
{
	char *src;

	return_if_fail(machine != NULL);

	src = (char *)&machine->stack[machine->sp >> 2];

	strncpy(dst, src, STRING_LEN);
	dst[STRING_LEN-1] = 0;
	vm_stack_grow(machine, STRING_LEN);
}

void vm_stack_peek_string(struct vm *machine, char dst[STRING_LEN])
{
	char *src;

	return_if_fail(machine != NULL);

	src = (char *)&machine->stack[machine->sp >> 2];

	strncpy(dst, src, STRING_LEN);
	dst[STRING_LEN-1] = 0;
}

void vm_stack_put_string(struct vm *machine, char src[STRING_LEN])
{
	char *dst;

	return_if_fail(machine != NULL);

	vm_stack_shrink(machine, STRING_LEN);
	dst = (char *)&machine->stack[machine->sp >> 2];

	strncpy(dst, src, STRING_LEN);
	dst[STRING_LEN-1] = 0;
}

void vm_stack_poke_string(struct vm *machine, char src[STRING_LEN])
{
	char *dst;

	return_if_fail(machine != NULL);

	dst = (char *)&machine->stack[machine->sp >> 2];

	strncpy(dst, src, STRING_LEN);
	dst[STRING_LEN-1] = 0;
}

static void vm_ent(struct vm *machine)
{
	vm_stack_put_int(machine, machine->fp);
	machine->fp = machine->sp;
}

static void vm_leave(struct vm *machine)
{
	machine->sp = machine->fp;
	machine->fp = vm_stack_get_int(machine);

	if (machine->pc == machine->stop_addr)
		return;

	machine->pc = vm_stack_peek_int(machine);

	if (machine->pc > machine->text_size)
		raise_exception(machine, EXCEPT_OO_TEXT);

	vm_stack_grow(machine, 4);
}

static void vm_grow(struct vm *machine)
{
	int32_t size;

	size = peek_operand(machine);
	machine->pc += 4;

	vm_stack_grow(machine, size);
}

static void vm_shrink(struct vm *machine)
{
	int32_t size;

	size = peek_operand(machine);
	machine->pc += 4;

	vm_stack_shrink(machine, size);
}

static void vm_push(struct vm *machine)
{
	uint32_t _const;
	float *fl;

	_const = peek_operand(machine);

	fl = (float *)&_const;

	machine->pc += 4;
	vm_stack_put_int(machine, _const);
}

static void vm_strpush(struct vm *machine)
{
	int32_t offs;
	char *src;

	offs = peek_operand(machine);
	machine->pc += 4;

	if (offs < 0 || offs > machine->ro_size)
		raise_exception(machine, EXCEPT_OO_RO);

	src = (char *)machine->ro + offs;

	vm_stack_put_string(machine, src);
}

static void vm_ioper(struct vm *machine, unsigned char opcode)
{
	int32_t op1, op2, res;

	op2 = vm_stack_get_int(machine);
	op1 = vm_stack_peek_int(machine);

	switch (opcode) {
	case ADD:
		res = op1 + op2;
		break;
	case SUB:
		res = op1 - op2;
		break;
	case MUL:
		res = op1 * op2;
		break;
	case DIV:
		if (op2 == 0) {
			raise_exception(machine, EXCEPT_ZERO_DIV);
			break;
		}
		res = op1 / op2;
		break;
	case MOD:
		res = op1 % op2;
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_OPCODE);
	}

	vm_stack_poke_int(machine, res);
}

static void vm_foper(struct vm *machine, unsigned char opcode)
{
	double op1, op2, res;

	op2 = vm_stack_get_float(machine);
	op1 = vm_stack_peek_float(machine);

	switch (opcode) {
	case FADD:
		res = op1 + op2;
		break;
	case FSUB:
		res = op1 - op2;
		break;
	case FMUL:
		res = op1 * op2;
		break;
	case FDIV:
		res = op1 / op2;
		break;
	case FMOD:
		res = fmodf(op1, op2);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_OPCODE);
	}

	vm_stack_poke_float(machine, res);
}

static void vm_jmp(struct vm *machine)
{
	int32_t offs;

	offs = peek_operand(machine);

	if (machine->pc+offs > machine->text_size)
		raise_exception(machine, EXCEPT_OO_TEXT);

	machine->pc += offs;
}

static void vm_cmp(struct vm *machine)
{
	int32_t op1, op2;

	op2 = vm_stack_get_int(machine);
	op1 = vm_stack_peek_int(machine);

	if (op1 > op2)
		vm_stack_poke_int(machine, 1);
	else if (op1 < op2)
		vm_stack_poke_int(machine, -1);
	else if (op1 == op2)
		vm_stack_poke_int(machine, 0);
}

static void vm_fcmp(struct vm *machine)
{
	double op1, op2;

	op2 = vm_stack_get_float(machine);
	op1 = vm_stack_peek_float(machine);

	if (op1 > op2)
		vm_stack_poke_int(machine, 1);
	else if (op1 < op2)
		vm_stack_poke_int(machine, -1);
	else if (op1 == op2)
		vm_stack_poke_int(machine, 0);
}

static void vm_strcmp(struct vm *machine)
{
	char op1[STRING_LEN];
	char op2[STRING_LEN];
	int32_t res;

	vm_stack_get_string(machine, op2);
	vm_stack_get_string(machine, op1);

	res = strncmp(op1, op2, sizeof(op1));
	vm_stack_put_int(machine, res);
}

static void vm_cond_jmp(struct vm *machine, unsigned char opcode)
{
	int32_t prev_res, offs;

	offs = peek_operand(machine);

	if (machine->pc+offs > machine->text_size)
		raise_exception(machine, EXCEPT_OO_TEXT);

	prev_res = vm_stack_get_int(machine);

	if (prev_res == 1 && (opcode == JG || opcode == JGE || opcode == JNEQ)) {
		machine->pc += offs;
	} else if (prev_res == -1 && (opcode == JL || opcode == JLE || opcode == JNEQ)) {
		machine->pc += offs;
	} else if (prev_res == 0 && (opcode == JGE || opcode == JLE || opcode == JEQ)) {
		machine->pc += offs;
	} else {
		machine->pc += 4;
	}
}

static void vm_call(struct vm *machine)
{
	uint32_t addr;

	addr = peek_operand(machine);
	machine->pc += 4;

	if (addr > machine->text_size)
		raise_exception(machine, EXCEPT_OO_TEXT);

	vm_stack_put_int(machine, machine->pc);
	machine->pc = addr;
}

static void check_bit_num(struct vm *machine, unsigned char data_type, unsigned char nbit)
{
	switch (data_type) {
	case BYTE:
	case UBYTE:
		if (nbit > 7)
			raise_exception(machine, EXCEPT_BIT_NUM);
		break;
	case SHORT:
	case USHORT:
		if (nbit > 16)
			raise_exception(machine, EXCEPT_BIT_NUM);
		break;
	case LONG:
	case ULONG:
	case FLOAT:
		if (nbit > 31)
			raise_exception(machine, EXCEPT_BIT_NUM);
		break;
	default:
		raise_exception(machine, EXCEPT_MEM_AREA_DTYPE);
	}
}

static uint32_t vm_read_inputs(struct vm *machine, uint32_t area, uint32_t idx)
{
	uint32_t offs, res, u32;
	uint16_t u16;
	unsigned char data_type, nbit;
	size_t size;

	data_type = GET_DATA_TYPE(area);
	offs = GET_OFFSET(area);

	switch (data_type) {
	case BYTE:
	case UBYTE:
		offs += idx;
		size = 1;
		break;
	case SHORT:
	case USHORT:
		offs += 2 * idx;
		size = 2;
		break;
	case LONG:
	case ULONG:
	case FLOAT:
		offs += 4 * idx;
		size = 4;
		break;
	default:
		size = 0;
		raise_exception(machine, EXCEPT_MEM_AREA_DTYPE);
	}

	check_area_bounds(machine, IAREA_OFFS, offs, size);

	switch (data_type) {
	case BYTE:
		res = (int8_t)buffer_get_u8_at(machine->inputs, offs);
		break;
	case UBYTE:
		res = buffer_get_u8_at(machine->inputs, offs);
		break;
	case SHORT:
		u16 = buffer_get_u16_at(machine->inputs, offs);
		u16 = adj_bo_u16(u16, machine->byte_order);
		res = (int16_t)u16;
		break;
	case USHORT:
		u16 = buffer_get_u16_at(machine->inputs, offs);
		u16 = adj_bo_u16(u16, machine->byte_order);
		res = u16;
		break;
	case LONG:
	case ULONG:
	case FLOAT:
		u32 = buffer_get_u32_at(machine->inputs, offs);
		u32 = adj_bo_u32(u32, machine->byte_order);
		res = u32;
		break;
	default:
		res = 0;
		raise_exception(machine, EXCEPT_MEM_AREA_DTYPE);
	}

	if (HAS_BIT(area)) {
		nbit = GET_BIT(area);
		check_bit_num(machine, data_type, nbit);

		if (CHECK_BIT(res, nbit))
			res = 1;
		else
			res = 0;
	}

	return res;
}

static void vm_write_outputs(struct vm *machine, uint32_t area, uint32_t data, uint32_t idx)
{
	uint32_t res, u32;
	uint16_t u16;
	int32_t offs;
	unsigned char data_type, nbit;
	size_t size;

	data_type = GET_DATA_TYPE(area);
	offs = GET_OFFSET(area);

	switch (data_type) {
	case BYTE:
	case UBYTE:
		offs += idx;
		size = 1;
		break;
	case SHORT:
	case USHORT:
		offs += 2 * idx;
		size = 2;
		break;
	case LONG:
	case ULONG:
	case FLOAT:
		offs += 4 * idx;
		size = 4;
		break;
	default:
		size = 0;
		raise_exception(machine, EXCEPT_MEM_AREA_DTYPE);
	}

	check_area_bounds(machine, QAREA_OFFS, offs, size);

	if (HAS_BIT(area)) {
		nbit = GET_BIT(area);
		check_bit_num(machine, data_type, nbit);

		switch (data_type) {
		case BYTE:
			res = (int8_t)buffer_get_u8_at(machine->outputs, offs); 
			break;
		case UBYTE:
			res = buffer_get_u8_at(machine->outputs, offs);
			break;
		case SHORT:
			u16 = buffer_get_u16_at(machine->outputs, offs);
			u16 = adj_bo_u16(u16, machine->byte_order);
			res = (int16_t)u16;
			break;
		case USHORT:
			u16 = buffer_get_u16_at(machine->outputs, offs);
			u16 = adj_bo_u16(u16, machine->byte_order);
			res = u16;
			break;
		case LONG:
		case ULONG:
		case FLOAT:
			u32 = buffer_get_u32_at(machine->outputs, offs);
			u32 = adj_bo_u32(u32, machine->byte_order);
			res = u32;
			break;
		default:
			res = 0;
			raise_exception(machine, EXCEPT_MEM_AREA_DTYPE);
		}

		if (data != 0)
			data = RAISE_BIT(res, nbit);
		else
			data = CLEAR_BIT(res, nbit);
	}

	switch (data_type) {
	case BYTE:
	case UBYTE:
		buffer_put_u8_at(machine->outputs, data, offs);
		break;
	case SHORT:
	case USHORT:
		u16 = adj_bo_u16(data, machine->byte_order);
		buffer_put_u16_at(machine->outputs, u16, offs);
		break;
	case LONG:
	case ULONG:
	case FLOAT:
		u32 = adj_bo_u32(data, machine->byte_order);
		buffer_put_u32_at(machine->outputs, u32, offs);
		break;
	default:
		raise_exception(machine, EXCEPT_MEM_AREA_DTYPE);
	}
}

static void vm_load(struct vm *machine)
{
	unsigned char area;
	int32_t offs;
	uint32_t datum;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	if (area == DATA_OFFS || area == FP_OFFS)
		check_area_bounds(machine, area, offs, 4);

	switch (area) {
	case DATA_OFFS:
		datum = machine->data[offs >> 2];
		break;
	case FP_OFFS:
		datum = machine->stack[(machine->fp+offs) >> 2];
		break;
	case IAREA_OFFS:
		datum = vm_read_inputs(machine, offs, 0);
		break;
	case QAREA_OFFS:
		raise_exception(machine, EXCEPT_READ_OUTPUTS);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	vm_stack_put_int(machine, datum);
}

static void vm_store(struct vm *machine, boolean_t leave)
{
	unsigned char area;
	int32_t offs;
	uint32_t datum;

	area = machine->text[machine->pc++];

	offs = peek_operand(machine);
	machine->pc += 4;

	datum = vm_stack_peek_int(machine);

	if (!leave)
		vm_stack_grow(machine, 4);

	if (area == DATA_OFFS || area == FP_OFFS)
		check_area_bounds(machine, area, offs, 4);

	switch (area) {
	case DATA_OFFS:
		machine->data[offs >> 2] = datum;
		break;
	case FP_OFFS:
		machine->stack[(machine->fp+offs) >> 2] = datum;
		break;
	case QAREA_OFFS:
		vm_write_outputs(machine, offs, datum, 0);
		break;
	case IAREA_OFFS:
		raise_exception(machine, EXCEPT_WRITE_INPUTS);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}
}

static void vm_aload(struct vm *machine)
{
	unsigned char area;
	int32_t offs, idx, dimen;
	uint32_t datum;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	idx = vm_stack_peek_int(machine);

	switch (area) {
	case DATA_OFFS:
		check_area_bounds(machine, area, offs, 4);
		dimen = machine->data[offs >> 2];
		break;
	case FP_OFFS:
		check_area_bounds(machine, area, offs, 4);
		dimen = machine->stack[(machine->fp+offs) >> 2];
		break;
	case IAREA_OFFS:
	case QAREA_OFFS:
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
		break;
	}

	if (area == DATA_OFFS || area == FP_OFFS) {
		if (idx < 0 || idx > dimen-1)
			raise_exception(machine, EXCEPT_OO_ARRAY);

		check_area_bounds(machine, area, offs+(idx+1)*4, 4);
	}

	switch (area) {
	case DATA_OFFS:
		datum = machine->data[(offs+(idx+1)*4) >> 2];
		break;
	case FP_OFFS:
		datum = machine->stack[(machine->fp+offs+(idx+1)*4) >> 2];
		break;
	case IAREA_OFFS:
		datum = vm_read_inputs(machine, offs, idx);
		break;
	case QAREA_OFFS:
		raise_exception(machine, EXCEPT_READ_OUTPUTS);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	vm_stack_poke_int(machine, datum);
}

static void vm_astore(struct vm *machine, boolean_t leave)
{
	unsigned char area;
	int32_t offs, idx, dimen;
	uint32_t datum;

	area = machine->text[machine->pc++];

	offs = peek_operand(machine);
	machine->pc += 4;

	idx = vm_stack_get_int(machine);
	datum = vm_stack_peek_int(machine);

	if (!leave)
		vm_stack_grow(machine, 4);

	switch (area) {
	case DATA_OFFS:
		check_area_bounds(machine, area, offs, 4);
		dimen = machine->data[offs >> 2];
		break;
	case FP_OFFS:
		check_area_bounds(machine, area, offs, 4);
		dimen = machine->stack[(machine->fp+offs) >> 2];
		break;
	case IAREA_OFFS:
	case QAREA_OFFS:
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	if (area == DATA_OFFS || area == FP_OFFS) {
		if (idx < 0 || idx > dimen-1)
			raise_exception(machine, EXCEPT_OO_ARRAY);

		check_area_bounds(machine, area, offs+(idx+1)*4, 4);
	}

	switch (area) {
	case DATA_OFFS:
		machine->data[(offs+(idx+1)*4) >> 2] = datum ;
		break;
	case FP_OFFS:
		machine->stack[(machine->fp+offs+(idx+1)*4) >> 2] = datum;
		break;
	case QAREA_OFFS:
		vm_write_outputs(machine, offs, datum, idx);
		break;
	case IAREA_OFFS:
		raise_exception(machine, EXCEPT_WRITE_INPUTS);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}
}

static void vm_ainit(struct vm *machine)
{
	unsigned char area;
	int32_t dimen, offs;

	area = machine->text[machine->pc++];

	offs = peek_operand(machine);
	machine->pc += 4;

	dimen = vm_stack_get_int(machine);
	check_area_bounds(machine, area, offs, 4);

	switch (area) {
	case DATA_OFFS:
		machine->data[offs >> 2] = dimen;
		break;
	case FP_OFFS:
		machine->stack[(machine->fp+offs) >> 2] = dimen;
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}
}

static void vm_strstore(struct vm *machine, boolean_t leave)
{
	int32_t offs;
	unsigned char area;
	char src[STRING_LEN], *dst;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	vm_stack_peek_string(machine, src);

	if (!leave)
		vm_stack_grow(machine, STRING_LEN);

	check_area_bounds(machine, area, offs, STRING_LEN);

	switch (area) {
	case DATA_OFFS:
		dst = (char *)&machine->data[offs >> 2];
		break;
	case FP_OFFS:
		dst = (char *)&machine->stack[(machine->fp+offs) >> 2];
		break;
	default:
		dst = NULL;
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	strncpy(dst, src, sizeof(src));
	dst[STRING_LEN-1] = 0;
}

static void vm_strload(struct vm *machine)
{
	int32_t offs;
	unsigned char area;
	char *src;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	check_area_bounds(machine, area, offs, STRING_LEN);

	switch (area) {
	case DATA_OFFS:
		src = (char *)&machine->data[offs >> 2];
		break;
	case FP_OFFS:
		src = (char *)&machine->stack[(machine->fp+offs) >> 2];
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	vm_stack_put_string(machine, src);
}

static void vm_astrload(struct vm *machine)
{
	int32_t offs, dimen, idx;
	unsigned char area;
	char *src;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	idx = vm_stack_get_int(machine);

	check_area_bounds(machine, area, offs, STRING_LEN);

	switch (area) {
	case DATA_OFFS:
		dimen = machine->data[offs >> 2];
		break;
	case FP_OFFS:
		dimen = machine->stack[(machine->fp+offs) >> 2];
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	if (idx < 0 || idx > dimen-1)
		raise_exception(machine, EXCEPT_OO_ARRAY);

	check_area_bounds(machine, area, offs+4+idx*STRING_LEN, STRING_LEN);

	switch (area) {
	case DATA_OFFS:
		src = (char *)&machine->data[(offs+4+idx*STRING_LEN) >> 2];
		break;
	case FP_OFFS:
		src = (char *)&machine->stack[(machine->fp+offs+4+idx*STRING_LEN) >> 2];
		break;
	default:
		src = NULL;
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	vm_stack_put_string(machine, src);
}

static void vm_astrstore(struct vm *machine, boolean_t leave)
{
	int32_t offs, dimen, idx;
	unsigned char area;
	char src[STRING_LEN], *dst;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	idx = vm_stack_get_int(machine);
	vm_stack_peek_string(machine, src);

	if (!leave)
		vm_stack_grow(machine, STRING_LEN);

	check_area_bounds(machine, area, offs, STRING_LEN);

	switch (area) {
	case DATA_OFFS:
		dimen = machine->data[offs >> 2];
		break;
	case FP_OFFS:
		dimen = machine->stack[(machine->fp+offs) >> 2];
		break;
	default:
		dimen = 0;
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	if (idx < 0 || idx > dimen-1)
		raise_exception(machine, EXCEPT_OO_ARRAY);

	check_area_bounds(machine, area, offs+4+idx*STRING_LEN, STRING_LEN);

	switch (area) {
	case DATA_OFFS:
		dst = (char *)&machine->data[(offs+4+idx*STRING_LEN) >> 2];
		break;
	case FP_OFFS:
		dst = (char *)&machine->stack[(machine->fp+offs+4+idx*STRING_LEN) >> 2];
		break;
	default:
		dst = NULL;
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}

	strncpy(dst, src, sizeof(src));
	dst[STRING_LEN-1] = 0;
}

static void vm_libcall(struct vm *machine)
{
	uint32_t ret_addr, sp, sysn;
	ret_t ret;

	ret_addr = vm_stack_get_int(machine);

	sysn = peek_operand(machine);
	machine->pc += 4;

	sp = machine->sp;
	ret = vm_libcall_call_by_num(machine, sysn);

	if (ret != ret_ok)
		raise_exception(machine, EXCEPT_LIBCALL);

	if (ret_addr > machine->text_size)
		raise_exception(machine, EXCEPT_OO_TEXT);

	machine->pc = ret_addr;
	machine->sp = sp;
}

static void vm_refput(struct vm *machine)
{
	int32_t offs;
	unsigned char area;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	check_area_bounds(machine, area, offs, 8);

	/* Put the reference base id and an absolute address. */
	switch (area) {
	case DATA_OFFS:
		vm_stack_put_int(machine, REF_DATA);
		vm_stack_put_int(machine, offs);
		break;
	case FP_OFFS:
		vm_stack_put_int(machine, REF_STACK);
		vm_stack_put_int(machine, machine->fp+offs);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}
}

static void vm_refcopy(struct vm *machine)
{
	int32_t offs;
	unsigned char area;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	check_area_bounds(machine, area, offs, 8);

	/* Copy the reference base id and an absolute address. */
	switch (area) {
	case FP_OFFS:
		vm_stack_put_int(machine, machine->stack[(machine->fp+offs+4) >> 2]);
		vm_stack_put_int(machine, machine->stack[(machine->fp+offs) >> 2]);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}
}

static void vm_get_refer_operands(struct vm *machine, unsigned char area, int32_t offs, uint32_t *abs_offs, uint32_t *ref_base)
{
	switch (area) {
	case FP_OFFS:
		check_area_bounds(machine, area, offs, 8);
		*abs_offs = machine->stack[(machine->fp+offs) >> 2];
		*ref_base = machine->stack[(machine->fp+offs+4) >> 2];
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_AREA);
	}
}

static void check_abs_offs(struct vm *machine, uint32_t ref_base, uint32_t abs_offs, size_t size)
{
	switch (ref_base) {
	case REF_DATA:
		if (abs_offs+size > machine->data_size)
			raise_exception(machine, EXCEPT_OO_DATA);
		break;
	case REF_STACK:
		if (abs_offs+size > machine->stack_size)
			raise_exception(machine, EXCEPT_OO_STACK);
		break;
	default:
		raise_exception(machine, EXCEPT_UKNOWN_REF_POINT);
	}
}

static void vm_refload(struct vm *machine)
{
	int32_t offs;
	uint32_t res, ref_base, abs_offs;
	unsigned char area;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	vm_get_refer_operands(machine, area, offs, &abs_offs, &ref_base);

	check_abs_offs(machine, ref_base, abs_offs, 4);

	switch (ref_base) {
	case REF_DATA:
		res = machine->data[abs_offs >> 2];
		break;
	case REF_STACK:
		res = machine->stack[abs_offs >> 2];
		break;
	default:
		raise_exception(machine, EXCEPT_UKNOWN_REF_POINT);
	}

	vm_stack_put_int(machine, res);
}

static void vm_refstore(struct vm *machine, boolean_t leave)
{
	int32_t offs;
	uint32_t datum, abs_offs, ref_base;
	unsigned char area;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	vm_get_refer_operands(machine, area, offs, &abs_offs, &ref_base);

	datum = vm_stack_peek_int(machine);
	if (!leave)
		vm_stack_grow(machine, 4);

	check_abs_offs(machine, ref_base, abs_offs, 4);

	switch (ref_base) {
	case REF_DATA:
		machine->data[abs_offs >> 2] = datum;
		break;
	case REF_STACK:
		machine->stack[abs_offs >> 2] = datum;
		break;
	default:
		raise_exception(machine, EXCEPT_UKNOWN_REF_POINT);
	}
}

static void vm_refstrload(struct vm *machine)
{
	int32_t offs;
	uint32_t ref_base, abs_offs;
	unsigned char area;
	char *src;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	vm_get_refer_operands(machine, area, offs, &abs_offs, &ref_base);

	check_abs_offs(machine, ref_base, abs_offs, STRING_LEN);

	switch (ref_base) {
	case REF_DATA:
		src = (char *)&machine->data[abs_offs >> 2];
		break;
	case REF_STACK:
		src = (char *)&machine->stack[abs_offs >> 2];
		break;
	default:
		raise_exception(machine, EXCEPT_UKNOWN_REF_POINT);
	}

	vm_stack_put_string(machine, src);
}

static void vm_refstrstore(struct vm *machine, boolean_t leave)
{
	int32_t offs;
	uint32_t abs_offs, ref_base;
	char src[STRING_LEN], *dst;
	unsigned char area;

	area = machine->text[machine->pc++];
	offs = peek_operand(machine);
	machine->pc += 4;

	vm_get_refer_operands(machine, area, offs, &abs_offs, &ref_base);

	vm_stack_peek_string(machine, src);
	if (!leave)
		vm_stack_grow(machine, STRING_LEN);

	check_abs_offs(machine, ref_base, abs_offs, STRING_LEN);

	switch (ref_base) {
	case REF_DATA:
		dst = (char *)&machine->data[abs_offs >> 2];
		break;
	case REF_STACK:
		dst = (char *)&machine->stack[abs_offs >> 2];
		break;
	default:
		dst = NULL;
		raise_exception(machine, EXCEPT_UKNOWN_REF_POINT);
	}

	strncpy(dst, src, sizeof(src));
	dst[STRING_LEN-1] = 0;
}

static void do_next_instr(struct vm *machine)
{
	unsigned char opcode;
	boolean_t leave;

	return_if_fail(machine != NULL);

	opcode = machine->text[machine->pc++];
	leave = FALSE;

	switch (opcode) {
	case ENT:
		vm_ent(machine);
		break;
	case LEAVE:
		vm_leave(machine);
		break;
	case GROW:
		vm_grow(machine);
		break;
	case SHRINK:
		vm_shrink(machine);
		break;
	case PUSH:
		vm_push(machine);
		break;
	case STRPUSH:
		vm_strpush(machine);
		break;
	case ADD:
	case SUB:
	case MUL:
	case DIV:
	case MOD:
		vm_ioper(machine, opcode);
		break;
	case FADD:
	case FSUB:
	case FMUL:
	case FDIV:
	case FMOD:
		vm_foper(machine, opcode);
		break;
	case JMP:
		vm_jmp(machine);
		break;
	case CMP:
		vm_cmp(machine);
		break;
	case FCMP:
		vm_fcmp(machine);
		break;
	case STRCMP:
		vm_strcmp(machine);
		break;
	case JG:
	case JGE:
	case JL:
	case JLE:
	case JEQ:
	case JNEQ:
		vm_cond_jmp(machine, opcode);
		break;
	case CALL:
		vm_call(machine);
		break;
	case LOAD:
		vm_load(machine);
		break;
	case STORE0:
		leave = TRUE;
	case STORE:
		vm_store(machine, leave);
		break;
	case ALOAD:
		vm_aload(machine);
		break;
	case ASTORE0:
		leave = TRUE;
	case ASTORE:
		vm_astore(machine, leave);
		break;
	case AINIT:
		vm_ainit(machine);
		break;
	case STRSTORE0:
		leave = TRUE;
	case STRSTORE:
		vm_strstore(machine, leave);
		break;
	case STRLOAD:
		vm_strload(machine);
		break;
	case ASTRLOAD:
		vm_astrload(machine);
		break;
	case ASTRSTORE0:
		leave = TRUE;
	case ASTRSTORE:
		vm_astrstore(machine, leave);
		break;
	case LIBCALL:
		vm_libcall(machine);
		break;
	case REFPUT:
		vm_refput(machine);
		break;
	case REFCOPY:
		vm_refcopy(machine);
		break;
	case REFLOAD:
		vm_refload(machine);
		break;
	case REFSTORE0:
		leave = TRUE;
	case REFSTORE:
		vm_refstore(machine, leave);
		break;
	case REFSTRLOAD:
		vm_refstrload(machine);
		break;
	case REFSTRSTORE0:
		leave = TRUE;
	case REFSTRSTORE:
		vm_refstrstore(machine, leave);
		break;
	default:
		raise_exception(machine, EXCEPT_UNKNOWN_OPCODE);
	}
}

void vm_exec_instr(struct vm *machine)
{
	return_if_fail(machine != NULL);

	machine->state = VM_RUNNING;

	do_next_instr(machine);

	if (machine->pc == machine->stop_addr)
		machine->state = VM_STOPPED;

	if (machine->vm_singlestep_callback != NULL)
		machine->vm_singlestep_callback(machine);
}

void vm_run(struct vm *machine)
{
	return_if_fail(machine != NULL);

	machine->state = VM_RUNNING;
	machine->pc = machine->start_addr;

	while(machine->state != VM_STOPPED) {
		do_next_instr(machine);
		if (machine->pc == machine->stop_addr)
			machine->state = VM_STOPPED;
	}
}

static int stop_addr_get(struct vm *machine, unsigned int *addr)
{
	struct sym_tb_entry s;
	struct buffer *sym_buf;
	unsigned int i;

	for (i = 0; i < machine->sym_tb_size; i++) {

		sym_buf = buffer_create_static_full(&machine->sym_tb[i],
						sizeof(struct sym_tb_entry));
		obj_get_sym_tb_entry(sym_buf, &s);
		buffer_free(sym_buf);

		if (s.st_shtype != SHT_TEXT)
			continue;

		if (s.st_offset != machine->start_addr)
			continue;

		*addr = machine->start_addr + s.st_size;

		return 0;
	}

	return -1;
}

static int start_addr_get(struct vm *machine, const char *sym_name,
			  unsigned int *addr)
{
	struct sym_tb_entry s;
	struct buffer *sym_buf;
	unsigned int i;

	for (i = 0; i < machine->sym_tb_size; i++) {

		sym_buf = buffer_create_static_full(&machine->sym_tb[i],
						sizeof(struct sym_tb_entry));
		obj_get_sym_tb_entry(sym_buf, &s);
		buffer_free(sym_buf);

		if (s.st_shtype != SHT_TEXT)
			continue;

		if (strcmp(sym_name, (char *)machine->string+s.st_name))
			continue;

		*addr = s.st_offset;
		return 0;
	}

	return -1;
}

ret_t vm_load_file(struct vm *machine, const char *file_name,
		   const char *sym_name, int lock_mem)
{
	struct buffer *obj_h_buf;
	struct obj_header obj_h;
	struct section_header sh;
	unsigned char *begin, *t;
	unsigned int i, addr;
	size_t len;
	ret_t ret;
	int fd, res;

	return_val_if_fail(machine != NULL, ret_err);
	return_val_if_fail(file_name != NULL, ret_err);

	ret = open_file_ro(file_name, &fd);
	if (ret != ret_ok)
		return ret;

	file_length(fd, &len);

	ret = mmap_file(fd, (void **)&begin, len);
	if (ret != ret_ok)
		return ret;

	close_file(&fd);

	obj_h_buf = buffer_create_static_full(begin, sizeof(struct obj_header));
	obj_get_obj_header(obj_h_buf, &obj_h);
	buffer_free(obj_h_buf);

	if (strncmp((char *)obj_h.o_magic, ".VM", 3)) {
		munmap_file((void **)begin, len);
		return ret_err;
	}

	if (obj_h.o_maj_vers != MAJOR_VERS || obj_h.o_min_vers != MINOR_VERS)
		xerror(1, "incompatible file version");

	for (t = begin + obj_h.o_shoff, i = 0; i < obj_h.o_shnum;
				t += obj_h.o_shentsize, i++) {
		struct buffer *sect_head_buf;

		sect_head_buf = buffer_create_static_full(t, sizeof(struct section_header));
		obj_get_section_header(sect_head_buf, &sh);
		buffer_free(sect_head_buf);

		switch (sh.sh_type) {
		case SHT_TEXT:
			machine->text = begin + sh.sh_off;
			machine->text_size = sh.sh_size;
			break;
		case SHT_INIT:
			machine->init = begin + sh.sh_off;
			machine->init_size = sh.sh_size;
			break;
		case SHT_DATA:
			machine->data_size = sh.sh_size;
			if (machine->data_size > 0)
				machine->data = xmalloc0(machine->data_size);
			break;
		case SHT_RO:
			machine->ro = begin + sh.sh_off;
			machine->ro_size = sh.sh_size;
			break;
		case SHT_SYMTAB:
			machine->sym_tb = (struct sym_tb_entry *)(begin + sh.sh_off);
			machine->sym_tb_size = sh.sh_size / sh.sh_entsize;
			break;
		case SHT_STRTAB:
			machine->string = begin + sh.sh_off;
			break;
		default:
			xerror(1, "unknown section header type");
		}
	}

	assert(machine->text != NULL);
	assert(machine->init != NULL);
	assert(machine->ro != NULL);
	assert(machine->sym_tb != NULL);
	assert(machine->string != NULL);

	machine->mmap_begin = begin;
	machine->mmap_size = len;

	machine->stack_size = STACK_SIZE * sizeof(uint32_t);
	machine->stack = xmalloc0(machine->stack_size);

	machine->sp = machine->stack_size;

	if (sym_name != NULL) {
		res = start_addr_get(machine, sym_name, &addr);
		if (res == -1) {
			munmap_file((void **)begin, len);
			return ret_err;
		}
		machine->start_addr = addr;
	} else {
		machine->start_addr = obj_h.o_entry;
	}

	res = stop_addr_get(machine, &addr);
	if (res == -1) {
		munmap_file((void **)begin, len);
		return ret_err;
	}

	machine->stop_addr = addr;
	
	if (lock_mem) {
		mem_touch_and_lock(machine->mmap_begin, machine->mmap_size);
		mem_touch_and_lock(machine->stack, machine->stack_size);
		if (machine->data != NULL)
			mem_touch_and_lock(machine->data, machine->data_size);
	}

	return ret_ok;
}

ret_t vm_unload_file(struct vm *machine)
{
	return_val_if_fail(machine, ret_invalid);

	xfree(machine->stack);

	if (machine->data != NULL) {
		xfree(machine->data);
		machine->data = NULL;
	}

	machine->stack = NULL;
	machine->data = NULL;
	machine->text = NULL;
	machine->init = NULL;
	machine->ro = NULL;
	machine->string = NULL;
	machine->sym_tb = NULL;

	munmap_file((void **)&machine->mmap_begin, machine->mmap_size);

	return ret_ok;
}

ret_t vm_load_lib(struct vm *machine, const char *path_to_lib)
{
	return_val_if_fail(machine != NULL, ret_err);
	return_val_if_fail(path_to_lib != NULL, ret_err);

	return vm_libcall_load(machine, path_to_lib);
}

ret_t vm_unload_lib(struct vm *machine)
{
	return_val_if_fail(machine != NULL, ret_err);

	return vm_libcall_unload(machine);
}

ret_t vm_prepare(struct vm *machine)
{
	unsigned char *text;

	return_val_if_fail(machine != NULL, ret_err);

	text = machine->text;

	machine->state = VM_RUNNING;
	machine->pc = 0;
	machine->text = machine->init;

	while(machine->pc != machine->init_size)
		do_next_instr(machine);

	machine->state = VM_STOPPED;
	machine->text = text;
	machine->pc = machine->start_addr;

	return ret_ok;
}

